(* Authors: Laura, Julian *)
open Sast
open Ast


(*environment*)
type symbol_table = {
    parent: symbol_table option;
    mutable variables: (Ast.typ * string) list;
    mutable declFuncs: (Ast.typ * string * (Ast.typ list)) list; (*list of usable functions: type, name, list of arguement types*)
}

type environment = {
    scope: symbol_table;
    func: Ast.typ; (*the scope's return type*)
    mutable returned: bool; (*does the scope have a return statement*)
    mutable pieceNames: string list;
    inLoop: bool; (*boolean that tells if you are in a loop. determines if break can be used*)
    
}

(*rewrite find variables
vdecl: find variable, if already found, raise error. otherwise, add to table*)

(*code for scope checking taken from edward's slides, can be changed*)
let rec var_local (scope: symbol_table) name = 
    List.exists(fun (_,s) -> s = name) scope.variables

let rec find_variable (scope: symbol_table) name =
    try List.find (fun (_, s) -> s = name) scope.variables 
    with Not_found ->
        match scope.parent with
        | Some(parent) -> find_variable parent name
        | None -> raise Not_found

let rec find_function (scope: symbol_table) name = 
    try List.find (fun (_, s, _) -> s = name) scope.declFuncs 
    with Not_found ->
        match scope.parent with
        | Some(parent) -> find_function parent name
        | None -> raise Not_found

(**)
let rec expr env = function 
    | Ast.BoolLiteral(b)    -> Ast.Bool, Sast.BoolLiteral(b)
    | Ast.IntLiteral (i)    -> Ast.Int, Sast.IntLiteral(i)
    | Ast.StringLiteral(s)  -> Ast.Str, Sast.StringLiteral(s)
    | Ast.Binop (e1,o,e2)   -> 
        let e1 = expr env e1 and e2 = expr env e2 in  (*evaluates left and right hand expressions*)
        let t1, _ = e1 and t2, _ = e2 in (*Obtain types of left and right expressions*)
        let opTypeFail expectedType =
            failwith (Ast.string_of_op o ^ " is only for defined for " ^ expectedType)
        in (match o with 
            | Ast.Add | Ast.Sub | Ast.Mult | Ast.Div | Ast.Mod ->
                if t1=t2 && t1=Ast.Int then Ast.Int, Sast.Binop(e1,o,e2)
                else opTypeFail "int"
            | Ast.Less | Ast.Greater | Ast.Leq | Ast.Geq ->
                if t1=t2 && t1=Ast.Int then Ast.Bool, Sast.Binop(e1,o,e2)
                else opTypeFail "int"
            | Equal | Neq ->
                if t1=t2 && (t1=Ast.Int || t1=Ast.Bool || t1=Ast.Piece || t1=Ast.Str) then Ast.Bool, Sast.Binop(e1,o,e2)
                else opTypeFail "int and int, or bool and bool, or piece and piece, or string and string"
            | And | Or ->
                if t1=t2 && t1=Ast.Bool then Ast.Bool, Sast.Binop(e1,o,e2)
                else opTypeFail "bool"
            | StrEqual -> 
                if t1=t2 && t1=Ast.Str then Ast.Bool, Sast.Binop(e1,o,e2)
                else opTypeFail "string"
        )
    | Ast.Unop (o,e) -> 
        let e = expr env e in (*evaluate e*)
        let t,_ = e in(*find type of expression*)
            if t=Ast.Int then t, Sast.Unop(o,e) else failwith("- only defined for int; " ^ Ast.string_of_typ t ^ " found.")
    | Ast.Parenth (e) -> 
        let e = expr env e in
        let t, _= e in
            t, Sast.Parenth(e)
    | Ast.Id(i) -> 
        let var = 
            try find_variable env.scope i
            with Not_found -> failwith("Uninitialised variable " ^ i ^ " referenced.")
        in let t,_ = var in 
        if t=Ast.Piece then failwith ("Piece types may not be explicitly called. Do you mean .type, .x, or .y?")
        else t, Sast.Id(i)
    | Ast.BoardAccess(x,y) -> 
        let varX = expr env x and varY=expr env y in 
        let tX,_=varX and tY,_=varY in
        if tX!=Ast.Int || tY!=Ast.Int then failwith ("Int types expected for board access. Other types found.")
        else Ast.Str, Sast.BoardAccess(varX,varY)
    | Ast.Access(v,f) -> 
        let var = 
            try find_variable env.scope v
            with Not_found -> failwith("Uninitialised variable " ^ v ^ " referenced.")
        in let t,_ = var in
        if t!=Ast.Piece then failwith (v ^ " is not of type Piece. Invalid command.")
        else if f="visible" then Ast.Bool, Sast.Access(v,f)
        else if f="type" then Ast.Str, Sast.Access(v,f)
        else if f="x" then Ast.Int, Sast.Access(v,f)
        else if f="y" then Ast.Int, Sast.Access(v,f)
        else failwith ("Invalid record name access.")
    | Ast.Empty -> Ast.Str, Sast.Empty
    | Ast.Rand(min, max) -> 
        let min= expr env min and max = expr env max in
        let tmin,_=min and tmax,_=max in
        if tmin=tmax && tmin=Ast.Int then Ast.Int, Sast.Rand(min,max)
        else failwith ("Parameters of random function must evaluate to integer values")

let rec typed_stmt_list env=  function
    []->[]
    |hd::tl ->
        let rec stmt env= function
        | Ast.VDecl (t,v,e) -> 
            if t=Ast.Piece then failwith ("Variable with piece type cannot be declared.")
            else if v="x" || v="y" || v="turn" then failwith ("Attempt to assign to reserved words x, y, or turn.")
            else if var_local env.scope v then failwith (v ^ " already declared in scope")
            else let e=expr env e in let t1,_=e in
            if t1!=t then failwith (Ast.string_of_typ t ^ " declared, " ^ Ast.string_of_typ t1 ^ " initialised in expression")
            else env.scope.variables<-(t,v)::env.scope.variables; Sast.VDecl(v, e)
        | Ast.Assign(v,e) -> 
        let var = 
            if v="x" || v="y" || v="turn" then failwith ("Attempt to assign to reserved words x, y, or turn.")
            else
            try find_variable env.scope v 
            with Not_found -> failwith("Uninitialised variable " ^ v ^ " referenced.")
        in let e = expr env e and t,_ = var
        in let t1,_ = e
        in
        if t=t1 then Sast.Assign(v, e)
        else failwith ("Attempt to assign " ^ Ast.string_of_typ t1 ^ " to variable declared as type " ^ string_of_typ t) 
        | Ast.Call (fname, argsList) -> 
            if fname = "break" && env.inLoop then Sast.Call(fname,[])
            else
            let f = 
            try find_function env.scope fname 
            with Not_found -> failwith ("Call to invalid/undeclared function.")
        in
        let  _,_,args=f in
            let numArgs = (List.length args) in
            if (List.length argsList)!=numArgs then failwith (string_of_int numArgs ^ " arguments expected, " ^ string_of_int (List.length argsList) ^ " found.")
            else 
                let rec checkList a ae = match a, ae with
                | [],[] -> []
                | _::_, [] | [],_::_-> failwith (":(")
                | hd::tl, hd1::tl1 -> (*hd has given expressions, hd1 is the type it should evaluate to*)
                    let helper hd hd1 = 
                        let e = expr env hd in let temptype,_=e in
                        if temptype!=hd1 then failwith ("Unexpected argument type.")
                        else e
                    in (helper hd hd1)::(checkList tl tl1)
                in Sast.Call(fname,checkList argsList args)
        | Ast.Expr(e) -> Sast.Expr(expr env e)
        | Ast.If(e, s1, s2) -> 
            let e = expr env e in 
            let t, _ = e in
            if t=Ast.Bool then 
                let scope1 = {env.scope with parent=Some(env.scope); variables=[];} and scope2={env.scope with parent=Some(env.scope); variables=[];} in
                let env1 = {env with scope=scope1} and env2 = {env with scope=scope2} in
                Sast.If(e, typed_stmt_list env1 s1, typed_stmt_list env2 s2)
            else failwith ("If predicate must be a boolean")
        | Ast.ForRow(v,e,s) -> 
            let e = expr env e in let t,_=e in
                if t!=Ast.Int then failwith ("Conditional expression must evaluate to an int value")
                else if var_local env.scope v then failwith (v ^ " already declared in local scope")
                else 
                    let scope' = {env.scope with parent=Some(env.scope); variables=[(Ast.Piece,v);]} in 
                    let env' = {env with scope=scope'; inLoop=true} in
                    Sast.ForRow(v, e, typed_stmt_list env' s)
        | Ast.ForCol(v,e,s) -> 
            let e = expr env e in let t,_=e in
                if t!=Ast.Int then failwith ("For expression must evaluate to an int value")
                else if var_local env.scope v then failwith (v ^ " already declared in local scope")
                else
                    let scope' = {env.scope with parent=Some(env.scope); variables=[(Ast.Piece,v)]} in 
                    let env' = {env with scope=scope'; inLoop=true} in
                    Sast.ForCol(v, e, typed_stmt_list env' s)
        | Ast.ForAll (v, s) -> 
            if var_local env.scope v then failwith (v ^ " already declared in local scope")
            else 
                let scope' = {env.scope with parent=Some(env.scope); variables=[(Ast.Piece,v)]} in 
                let env' = {env with scope=scope'; inLoop=true} in
                Sast.ForAll(v,typed_stmt_list env' s)
        | Ast.ForSurrounding(v, e1, e2,s) ->
            let e1=expr env e1 and e2=expr env e2 in
            let t1,_=e1 and t2,_=e2 in
            if t1!=Ast.Int || t2!=Ast.Int then failwith("For loop: expressions must evaluate to int values")
            else if var_local env.scope v then failwith (v ^ " already declared in local scope")
            else
                let scope' = {env.scope with parent=Some(env.scope); variables=[(Ast.Piece,v)]} in 
                let env' = {env with scope=scope'; inLoop=true} in
                Sast.ForSurrounding(v, e1, e2, typed_stmt_list env' s)
        | Ast.ForMinMax(v,e1,e2,s) ->
            let e1=expr env e1 and e2=expr env e2 in
            let t1,_=e1 and t2,_=e2 in
            if t1!=Ast.Int || t2!=Ast.Int then failwith("For loop: expressions must evaluate to int values")
            else if var_local env.scope v then failwith (v ^ " already declared in local scope")
            else 
                let scope' = {env.scope with parent=Some(env.scope); variables=[(Ast.Int,v)]} in 
                let env' = {env with scope=scope'; inLoop=true} in
                Sast.ForMinMax(v, e1, e2, typed_stmt_list env' s)
        | Ast.ForNested(v1,e11,e12,v2,e21,e22,s) ->
            let e11=expr env e11 and e12=expr env e12 and e21=expr env e21 and e22=expr env e22 in
            let t11,_=e11 and t12,_=e12 and t21,_=e21 and t22,_=e22 in
            if t11!=Ast.Int || t12!=Ast.Int || t21!=Ast.Int || t22!=Ast.Int then failwith("For loop: expressions must evaluate to int values")
            else if var_local env.scope v1 then failwith (v1 ^ " already declared in local scope")
            else if var_local env.scope v2 then failwith (v2 ^ " already declared in local scope")
            else
                let scope' = {env.scope with parent=Some(env.scope); variables=[(Ast.Int,v1);(Ast.Int,v2)]} in 
                let env' = {env with scope=scope'; inLoop=true} in
                Sast.ForNested(v1, e11, e12, v2, e21, e22, typed_stmt_list env' s)
        | Ast.ForSurroundingCoords(v1,v2,e1,e2,s) ->
            let e1=expr env e1 and e2=expr env e2 in
            let t1,_=e1 and t2,_=e2 in
            if t1!=Ast.Int || t2!=Ast.Int then failwith("For loop: expressions must evaluate to int values")
            else if var_local env.scope v1 then failwith (v1 ^ " already declared in local scope")
            else if var_local env.scope v2 then failwith (v2 ^ " already declared in local scope")
            else
                let scope' = {env.scope with parent=Some(env.scope); variables=[(Ast.Int,v1);(Ast.Int,v2)]} in 
                let env' = {env with scope=scope'; inLoop=true} in
                Sast.ForSurroundingCoords(v1, v2, e1, e2, typed_stmt_list env' s)
        | Ast.ForBoard(v1,v2,s) -> 
            if var_local env.scope v1 then failwith (v1 ^ " already declared in local scope")
            else if var_local env.scope v2 then failwith (v2 ^ " already declared in local scope")
            else
                let scope' = {env.scope with parent=Some(env.scope); variables=[(Ast.Int,v1);(Ast.Int,v2)]} in 
                let env' = {env with scope=scope'; inLoop=true} in
                Sast.ForBoard(v1, v2, typed_stmt_list env' s)
        | Ast.Return(e) ->
            let e = expr env e in let t,_= e in 
            if env.func=Ast.Void then failwith ("Return statement not allowed in void functions.") 
            else if t=env.func then (env.returned<-true; Sast.Return(e))
            else failwith("Return type does not match declared function return type.")
        in stmt env hd::typed_stmt_list env tl

let semcheck (board, players, turnOrder, pieces, setup, conditions) =
    (* check players *)
    let checkPlayers =
        (* check for at least one player *)
        (* check for duplicate player names, raise exception if found *)
        let reportError except1 except2 lst =
            if List.length lst < 1 then raise (Failure except1);
            let rec helper = function
                  n1 :: n2 :: _ when n1 = n2 -> raise (Failure (except2 n1))
                | _ :: t -> helper t
                | [] -> ()
            in helper(List.sort compare lst)
        in 
        reportError "no players" (fun n -> "duplicate player name " ^ n) 
        players
    in 
    checkPlayers;

    (* check turnOrder *)
    let checkTurnOrder =
        (* check for at least one player in turnOrder *)
        (* check for any undefined players in turnOrder *)
        let reportError except1 except2 lst1 lst2 =
            if List.length lst2 < 1 then raise (Failure except1);
            let rec helper lst p = match lst with
                  [] -> 0
                | hd :: tl when hd = p -> 1
                | _ :: tl -> helper tl p
            in 
            let rec check = function
                  [] -> 0
                | hd :: tl -> helper lst1 hd + check tl
            in 
            if check lst2 < List.length lst2 then raise (Failure except2) 
        in 
        reportError "no players in turnOrder" 
        "undefined players in turnOrder" players turnOrder
    in
    checkTurnOrder;


    let env = {scope = { variables=[(Ast.Int, "width");(Ast.Int, "height");]; parent=None; 
                declFuncs= [(Ast.Void,"setVisibility",[Ast.Int;Ast.Int;Ast.Int]);(Ast.Int,"rand",[Ast.Int;Ast.Int]);(Ast.Void,"print",[Ast.Str]);(Ast.Void, "place",[Ast.Str;Ast.Int;Ast.Int]);
        (Ast.Void,"removePiece",[Ast.Int;Ast.Int]); (Ast.Void, "click", [Ast.Int;Ast.Int]);(Ast.Int,"random",[Ast.Int;Ast.Int]) ]; };
        func=Ast.Void;returned=false; pieceNames=[]; inLoop=false;} in (*global environment includes print, place, removePiece, click functions. return type is void. no variables, no parent*)

    let checkBoard env board = 
    let x = expr env board.x and y = expr env board.y and bg=expr env board.bg in
    let x_type,_ = x and y_type,_=y and bg_type,_=bg in
    if x_type!=Ast.Int || y_type!=Ast.Int then failwith ("Board Dimensions must be an integer")
    else if bg_type!=Ast.Str then failwith ("Board image must be a string")
    else {x=x; y=y; img=bg;} in

    let rec checkPieces env=function
    []->[]
    | hd::tl -> 
        let onTurnScope = {parent=Some(env.scope); variables=[(Ast.Str, "turn");(Ast.Int,"x");(Ast.Int,"y")]; declFuncs=[(Ast.Void, "remove",[]);(Ast.Void, "visible",[Ast.Bool]);
        (Ast.Void,"endTurn",[]);(Ast.Void, "changeType", [Ast.Str])]} 
        and onClickScope = {parent=Some(env.scope); variables=[(Ast.Str, "turn");(Ast.Int,"x");(Ast.Int,"y")]; declFuncs=[(Ast.Void, "remove",[]);(Ast.Void, "visible",[Ast.Bool]);
        (Ast.Void,"endTurn",[]);(Ast.Void, "changeType", [Ast.Str])]} 
        in
        let onTurnEnv = {env with scope=onTurnScope} and onClickEnv = {env with scope=onClickScope} in
        let checkPiece env piece =   
            if (List.exists (fun (s) -> s = piece.name) env.pieceNames) then failwith ("Duplicate piece names found") 
            else env.pieceNames<- piece.name::env.pieceNames; {name=piece.name;img=piece.img;onTurn_typed=(typed_stmt_list onTurnEnv piece.onTurn); onClick_typed=(typed_stmt_list onClickEnv piece.onClick);}
        in checkPiece env hd::checkPieces env tl
    in  

    let winScope = {env.scope with parent=Some(env.scope); variables=[(Ast.Str, "turn");]; } and loseScope = {env.scope with parent=Some(env.scope); variables=[]; } and 
        drawScope = {env.scope with parent=Some(env.scope); variables=[(Ast.Str, "turn");]; } in
    let winEnv = {env with scope=winScope;func=Ast.Bool; returned=false} and loseEnv = {env with scope=loseScope;func=Ast.Bool;returned=false} and drawEnv={env with scope=drawScope; func=Ast.Bool;returned=false} in


    let scopeSetup = {env.scope with parent=Some(env.scope); variables=[]; } in 
    let envSetup = {env with scope=scopeSetup} in

    let board = checkBoard env board in 
    let setup = typed_stmt_list envSetup setup in
    let pieces = checkPieces env pieces in


    let l=conditions.l and w=conditions.w and d=conditions.d in 
    let conditions = {
        win_typed = typed_stmt_list winEnv conditions.win;
        lose_typed = typed_stmt_list loseEnv conditions.lose;
        draw_typed = typed_stmt_list drawEnv conditions.draw;
    } in

    
(*check players and turnorder need to be added*)

    if (winEnv.returned)=false && w then failwith("Win condition requires a return statement.")
    else if (loseEnv.returned)=false && l then failwith("Lose condition requires a return statement.") 
    else if (drawEnv.returned)=false && d then failwith("Draw condition requires a return statement.")
    else
        board, players, turnOrder, pieces, setup, conditions;
